"""Bayesian inference of a sparse inverse covariance matrix.

Based on
https://github.com/probml/pmtksupport/blob/master/GGM-GWishart/GWishartScore.m
"""
from functools import partial

import numpy as np
from numpy import binary_repr
from scipy import linalg
from scipy.optimize import minimize
from scipy.special import comb
from sklearn.covariance import empirical_covariance
from sklearn.covariance.empirical_covariance_ import log_likelihood
from sklearn.linear_model import LassoLars
from sklearn.utils import Bunch, check_array
from sklearn.utils.extmath import fast_logdet

from regain.covariance.graph_lasso_ import GraphLasso, graph_lasso


def mk_all_ugs(n_dim):
    """Utility for generating all possible """
    nedges = int(comb(n_dim, 2))
    m = 2 ** nedges

    ind = np.array([list(binary_repr(x, width=len(binary_repr(m-1))))
                    for x in range(m)]).astype(int)
    ord = np.argsort(ind.sum(axis=1))
    ind = ind[ord]

    ut = np.triu(np.ones((n_dim, n_dim)), 1) > 0

    Gs = []
    for i in range(m):
        G = np.zeros((n_dim, n_dim))
        G[ut] = ind[i]
        G = (G + G.T) > 0
        Gs.append(G)
    return Gs


def markov_blankets(graphs, boolean=True, tol=1e-8, unique=False):
    """For each variable, list each of its markov blankets in graphs."""
    m_blankets = [np.array([G[i] for G in graphs]) for i in
                  range(graphs[0].shape[0])]
    for i, mb in enumerate(m_blankets):
        mb[:, i] = 0
        mb[np.abs(mb) < tol] = 0
    if boolean:
        m_blankets = [mb != 0 for mb in m_blankets]
    if unique:
        # discard same blankets
        # works with numpy >= 1.13.0
        m_blankets = np.unique(np.array(m_blankets), axis=1)
    return m_blankets


def score_blankets(blankets, X, alphas=(0.01, 0.5, 1)):
    """Score the markov blankets for node i.

    Restrict data to markov blanket, use as y (the variable to regress)
    the node i.
    """
    scores_all = []
    for i, mb_i in enumerate(blankets):
        scores = []
        for j, mb in enumerate(mb_i):
            X_mb = X[:, mb]
            if X_mb.shape[1] < 1:
                X_mb = np.zeros((X.shape[0], 1))

            y_mb = X[:, i]
            score = np.sum([LassoLars(alpha=alpha).fit(
                X_mb, y_mb).score(X_mb, y_mb) for alpha in alphas])

            scores.append(score)
        scores_all.append(scores)

    ss = np.exp(np.array(scores_all))
    normalized_scores = ss / np.repeat(ss.sum(axis=1)[:, None],
                                       ss.shape[1], axis=1)
    return normalized_scores


def get_graphs(blankets, scores, n_dim, n_resampling=200):
    idx = np.array([np.random.choice(
        scores.shape[1], p=scores[i], size=n_resampling)
        for i in range(n_dim)])

    graphs_ = np.array([blankets[i][idx[i]] for i in
                        range(n_dim)]).transpose(1, 0, 2)
    # symmetrise with AND operator -> product
    graphs = np.array([g * g.T for g in graphs_])
    return graphs


def covsel(x, p, nonZero, C):
    """Objective and gradient for MLE of precision given empirical covariance.

    nonZero is a list of non-zero upper triangle precision matrix entries.
    Based on sparse GGM estimation code by Mark Schmidt.
    """
    X = np.zeros((p, p))
    X[nonZero] = x   # fill the diagonal and upper triangle
    X += np.triu(X, 1).T  # fill the lower triangle

    # Fast Way to compute -logdet(X) + tr(X*C)
    # f = -2*sum(log(diag(R))) + sum(sum(C.*X)) + (lambda/2)*sum(X(:).^2);
    f = - fast_logdet(X) + np.sum(C * X)

    if f < np.inf:
        g = C - linalg.pinvh(X)
        g += np.tril(g, -1).T  # add contribution from lower to upper triangle
        g = g[nonZero]
    else:
        g = 0
    return f, g


def precision_selection(G, n_dim, C):
    """Need to specify G"""
    function = partial(covsel, p=n_dim, nonZero=G, C=C)

    x0 = np.eye(n_dim)
    min_obj = minimize(function, x0[G], jac=True)
    x = min_obj.x

    XX = np.zeros((n_dim, n_dim))
    XX[G] = x
    XX += np.triu(XX, 1).T
    return XX


def GWishartFit(X, G, GWprior, mode='covsel'):
    n_samples, n_dim = X.shape

    d0 = GWprior.d0
    S0 = GWprior.S0

    # check prior size violations
    if G.shape[0] != n_dim or G.shape[1] != n_dim:
        raise ValueError('G must be p-by-p, with p dimensions X')
    if S0.shape[0] != n_dim or S0.shape[1] != n_dim:
        raise ValueError('GWprior.S0 must be p-by-p, with p dimensions X')

    # compute posterior scatter matrix
    dn = n_samples + d0

    # X'*X - but I dont assume X to be centered
    emp_cov = empirical_covariance(X)
    S = n_samples * emp_cov
    C = (S + S0) / (dn - 2)

    if mode == 'covsel':
        precision = precision_selection(G, n_dim, C)
    else:
        # use graph_lasso
        # convert G to alpha
        alpha = np.zeros_like(G, dtype=float)
        alpha[~(G + G.T)] = np.inf
        precision = graph_lasso(emp_cov=C, alpha=alpha)[0]

    return precision, S


def compute_score(X, G, P, S, GWprior=None, score_method='bic'):
    n_samples, n_dim = X.shape

    d0 = GWprior.d0
    S0 = GWprior.S0

    # check prior size violations
    if S0.shape[0] != n_dim or S0.shape[1] != n_dim:
        raise ValueError('GWprior.S0 must be p-by-p, with p dimensions X')

    dn = n_samples + d0
    # C = (S + S0) / (dn - 2)

    # % need logdetP and invP
    # es, Q = np.linalg.eigh(x)
    # Inv = np.linalg.multi_dot((Q, np.diag(1. / es), Q.T))
    U, s, Vh = linalg.svd(P)

    # check
    invP = np.linalg.multi_dot((Vh.T, np.diag(1. / s), U.T))
    logdetP = np.sum(np.log(s))

    # % compute loglik
    loglik = n_samples * log_likelihood(S / n_samples, P)

    num_edges = np.triu(G, 1).sum()
    dof = num_edges + n_dim

    # pcor = cov2cor(P);

    # % the posterior Sn parameter
    # Sn = (dn - 2) * invP
    logh = (dn - 2) / 2. * (n_dim + logdetP)

    # find full param set V
    Vi, Vj = np.nonzero(np.triu(G))

    # to be the same as matlab
    idx = np.argsort(Vj)
    Vi, Vj = Vi[idx], Vj[idx]

    GWpost = Bunch()
    GWpost.Sn = S + S0
    # GWpost.C = C
    GWpost.dn = dn
    GWpost.P = P
    GWpost.num_edges = num_edges
    GWpost.dof = dof
    GWpost.logdetP = logdetP
    GWpost.loglik = loglik

    if score_method == 'bic':
        score = loglik - dof * np.log(n_samples) / 2 if n_samples > 0 else 0

    elif score_method == 'diaglaplace':
        # Diagonal hessian laplace approximation
        diagH = np.zeros(dof)
        for e1 in range(dof):
            # e2 = e1

            M1 = np.zeros((n_dim, n_dim))
            # M2 = M1.copy()

            nz1 = [Vi[e1], Vj[e1]]
            # nz2 = [Vi[e2], Vj[e2]]
            M1[:, nz1] = invP[:, [Vj[e1], Vi[e1]]]
            # M2[:, nz2] = invP[:, [Vj[e2], Vi[e2]]]

            # A = M1[nz2][:, nz1]
            # B = M2[nz1][:, nz2]
            A = M1[nz1][:, nz1]
            B = A

            tmp2 = A[0, :].dot(B[:, 0]) + A[1, :].dot(B[:, 1])

            diagH[e1] = -(dn-2) * tmp2 / 2
            # diagH(e1) = -(dn-2) * trace(M1(nz2,nz1)*M2(nz1,nz2))/2;

        logdetHdiag = sum(np.log(-diagH))
        lognormconst = dof * np.log(2*np.pi)/2 + logh - logdetHdiag / 2.
        score = lognormconst - GWprior.lognormconst - \
            n_samples * n_dim * np.log(2 * np.pi) / 2
        GWpost.lognormconst = lognormconst

    elif score_method == 'laplace':
        # Full laplace approximation
        H = np.empty((dof, dof))
        for e1 in range(dof):
            # nz1 = [Vi[e1], Vj[e1]]
            i, j = Vi[e1], Vj[e1]

            for e2 in range(e1, dof):
                # nz2 = [Vi[e2], Vj[e2]]
                l, m = Vi[e2], Vj[e2]
                # A = invP[nz2][:, [Vj[e1], Vi[e1]]]
                # B = invP[nz1][:, [Vj[e2], Vi[e2]]]
                A = invP[[l, m]][:, [j, i]]
                B = invP[[i, j]][:, [m, l]]

                # tmp2 = A[0, :].dot(B[:, 0]) + A[1, :].dot(B[:, 1])
                # tmp2 = np.trace(A.dot(B))
                tmp2 = (A * B.T).sum()
                H[e2, e1] = H[e1, e2] = -(dn-2) * tmp2 / 2.

        # neg Hessian will be posdef
        logdetH = 2*sum(np.log(np.diag(linalg.cholesky(-H))))
        lognormconst = dof * np.log(2*np.pi)/2 + logh - logdetH / 2.
        score = lognormconst - GWprior.lognormconst - \
            n_samples * n_dim * np.log(2 * np.pi) / 2
        GWpost.lognormconst = lognormconst

    GWpost.score = score
    return GWpost


def GWishartScore(X, G, d0=3, S0=None, score_method='bic', mode='covsel'):
    # %Initialize GW prior structure
    n_samples, n_dim = X.shape
    if S0 is None:
        S0 = (d0 - 2) * np.diag(np.diag(np.cov(X,bias=True)))

    GWprior = Bunch(d0=d0, S0=S0, lognormconst=0, lognormconstDiag=0)

    # %If method is laplace, compute laplace approximation to denominator
    if score_method in ['laplace', 'diaglaplace']:
        noData = np.zeros((0, n_dim))

        P0, S_noData = GWishartFit(noData, G, GWprior)
        GWtemp = compute_score(noData, G, P0, S_noData, GWprior=GWprior,
                               score_method=score_method)
        GWprior.lognormconst = GWtemp.lognormconst

    # Compute the map precision matrix P
    P, S = GWishartFit(X, G, GWprior, mode=mode)

    # Compute the score
    GWpost = compute_score(X, G, P, S, GWprior, score_method)
    return GWpost


def bayesian_graph_lasso(
        X, tol=1e-8, alphas=None, n_resampling=200, mode='gl', top_n=1,
        scoring='diaglaplace', assume_centered=False, max_iter=100):
    """Bayesian graphical lasso."""
    n_samples, n_dim = X.shape

    if alphas is None:
        alphas = np.logspace(-2, 0, 20)

    # get a series of Markov blankets for vaiours alphas
    mdl = GraphLasso(assume_centered=assume_centered, tol=tol,
                     max_iter=max_iter, verbose=False)
    precisions = [
        mdl.set_params(alpha=a).fit(X).precision_
        for a in alphas]
    mblankets = markov_blankets(precisions, tol=tol, unique=True)

    normalized_scores = score_blankets(mblankets, X=X, alphas=[0.01, 0.5, 1])

    graphs = get_graphs(mblankets, normalized_scores, n_dim=n_dim,
                        n_resampling=n_resampling)

    nonzeros_all = [np.triu(g, 1) + np.eye(n_dim, dtype=bool) for g in graphs]

    # Roverato'02: convert from HIW to G-Wishart (delta + |V| - 1)
    d0 = 3 + n_dim - 1
    S0 = np.eye(n_dim)  # same as Roverato'02

    # Find non-zero elements of upper triangle of G
    # make sure diagonal is non-zero
    # G = nonzeros_all[1] # probably can discard if all zeros?
    res = [GWishartScore(X, G, d0=d0, S0=S0, mode=mode, score_method=scoring)
           for G in nonzeros_all]

    top_n = [x.P for x in sorted(res, key=lambda x: x.score)[::-1][:top_n]]
    return np.mean(top_n, axis=0)


class BayesianGraphLasso(GraphLasso):
    """Sparse inverse covariance estimation with an l1-penalized estimator.

    Parameters
    ----------
    alpha : positive float, default 0.01
        The regularization parameter: the higher alpha, the more
        regularization, the sparser the inverse covariance.

    rho : positive float, default 1
        Augmented Lagrangian parameter.

    over_relax : positive float, deafult 1
        Over-relaxation parameter (typically between 1.0 and 1.8).

    tol : positive float, default 1e-4
        Absolute tolerance to declare convergence.

    rtol : positive float, default 1e-4
        Relative tolerance to declare convergence.

    max_iter : integer, default 100
        The maximum number of iterations.

    verbose : boolean, default False
        If verbose is True, the objective function, rnorm and snorm are
        printed at each iteration.

    assume_centered : boolean, default False
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    update_rho_options : dict, default None
        Options for the update of rho. See `update_rho` function for details.

    compute_objective : boolean, default True
        Choose if compute the objective function during iterations
        (only useful if `verbose=True`).

    mode : {'admm'}, default 'admm'
        Minimisation algorithm. At the moment, only 'admm' is available,
        so this is ignored.

    Attributes
    ----------
    covariance_ : array-like, shape (n_features, n_features)
        Estimated covariance matrix

    precision_ : array-like, shape (n_features, n_features)
        Estimated pseudo inverse matrix.

    n_iter_ : int
        Number of iterations run.

    """

    def __init__(self, alpha=0.01, max_iter=100, alphas=None, n_resampling=200,
                 tol=1e-4, verbose=False, assume_centered=False, mode='gl',
                 scoring='diaglaplace', top_n=1):
        super(GraphLasso, self).__init__(
            alpha=alpha, tol=tol, max_iter=max_iter, verbose=verbose,
            assume_centered=assume_centered, mode=mode)
        self.alphas = alphas
        self.n_resampling = n_resampling
        self.scoring = scoring
        self.top_n = top_n

    def fit(self, X, y=None):
        """Fit the GraphLasso model to X.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data from which to compute the covariance estimate
        y : (ignored)

        """
        # Covariance does not make sense for a single feature
        X = check_array(X, ensure_min_features=2, ensure_min_samples=2,
                        estimator=self)

        if self.alphas is None:
            self.alphas = np.logspace(-2, 0, 20)

        self.precision_ = bayesian_graph_lasso(
            X, tol=self.tol, alphas=self.alphas,
            n_resampling=self.n_resampling, mode=self.mode,
            scoring=self.scoring, assume_centered=self.assume_centered,
            max_iter=self.max_iter, top_n=self.top_n)
        return self



from __future__ import print_function

from functools import partial

import numpy as np
from scipy import linalg
from sklearn.datasets.base import Bunch
from sklearn.metrics.pairwise import rbf_kernel
from sklearn.gaussian_process import kernels
from sklearn.utils.validation import check_X_y

from regain.bayesian import stats
from regain.bayesian.gaussian_process_ import sample as sample_gp
from regain.bayesian.sampling import elliptical_slice, sample_hyper_kernel
from regain.covariance.time_graph_lasso_ import TimeGraphLasso

from regainpr.bayesian.sampling import sample_ell


def GWP_construct(umat, L, uut=None):
    """Build the sample from the GWP.

    Optimised with uut:
    uut = np.array([u.dot(u.T) for u in umat.T])
    """

    if uut is None:
        v, p, n = umat.shape
        M = np.zeros((p, p, n))
        for i in range(n):
            for j in range(v):
                Lu = L.dot(umat[j, :, i])
                LuuL = Lu[:, None].dot(Lu[None, :])
                M[..., i] += LuuL

    else:
        M = np.array(
            [np.linalg.multi_dot((L, uu_i, L.T)) for uu_i in uut]).transpose()

    # assert np.allclose(N, M)
    return M


def fit(
        lp, var_prop, mu_prior, var_prior, var_Lprop, mu_Lprior, var_Lprior,
        kern, p, nu=None, t=None, n_iter=500, verbose=False, likelihood=None):
    """Sample the parameters of kernel and lower Cholesky.
 
    Parameters
    ----------
    lp, var_prop, mu_prior, var_prior : sampling kernel hyperparameters
    var_Lprop, mu_Lprior, var_Lprior : sampling the elements in the matrix L
    kern : function
        Function to compute the kernel. Should return a normalised square
        matrix.
    p : int
        Dimension of the problem.
    nu : int
        Degrees of freedom, usually nu = p + 1.
    """
    if nu is None:
        nu = p + 1

    K = kern(t[:, None], inverse_width=lp)
    umat = sample_gp(K, nu=nu, p=p)

    L = np.tril(np.random.randn(p, p))
    # Cholesky factor for the scale matrix V = LL^{\top}

    D = GWP_construct(np.real(umat), L)
    cur_log_like = likelihood(D)

    # The struct ff is formed for the ESS procedure
    ff = Bunch(xx=umat, V=D, L=L, uut=np.array([u.dot(u.T) for u in umat.T]))

    Ltau = L[np.tril_indices_from(L)]
    L__ = np.zeros((p, p))
    L__[np.tril_indices_from(L__)] = Ltau

    samples_u = []  # np.zeros((uvec.size, niters));
    loglikes = np.zeros(n_iter)
    lps = np.zeros(n_iter)
    Ls = []

    for i in range(n_iter):
        # We first do ESS to obtain a new sample for u
        if verbose:
            print(i, "%.3e" % cur_log_like, end='\r')

        ff, cur_log_like = elliptical_slice(
            ff, umat, cur_log_like, likelihood=likelihood)

        # We now do MH for sampling the hyperparameter of the kernel
        lp, accept = sample_hyper_kernel(
            lp, var_prop, t, ff.xx, kern, mu_prior, var_prior)

        uut = ff.uut
        # We now do MH for sampling the elements in the matrix L
        Ltau = sample_ell(
            Ltau, var_Lprop, ff.xx, mu_Lprior, var_Lprior, uut=uut,
            likelihood=likelihood)

        L__[np.tril_indices_from(L__)] = Ltau
        ff['L'] = L__

        if accept:
            # new kernel parameter, recompute
            K = kern(t[:, None], inverse_width=lp)
            while True:
                try:
                    umat = sample_gp(K, nu=nu, p=p)
                    break
                except:
                    K += 1e-8 * np.eye(t.size)

            uut = np.array([u.dot(u.T) for u in umat.T])

        samples_u.append(ff.xx)
        loglikes[i] = cur_log_like
        lps[i] = lp
        Ls.append(L__)
    return samples_u, loglikes, lps, Ls


def predict(t_test, t_train, u_map, L_map, kern, inverse_width_map):
    """Predict covariance matrix for t_test.

    Parameters
    ----------
    t_{test, train} : ndarray
        Test and train time points.
    u_map : type
        MAP estimate of u parameter.
    L_map : type
        MAP estimate of L parameter.
    kern : type
        Kernel function.
    inverse_width_map : float
        MAP estimate of inverse_width, the kernel parameter.

    Returns
    -------
    ndarray, shape (p, p, t_test.size)
        List of covariances at t_test.

    """
    # Compute the mean for ustar for test data
    KB = kern(t_train[:, None], inverse_width=inverse_width_map)
    A = kern(
        t_test[:, None], t_train[:, None], inverse_width=inverse_width_map)
    invKB = linalg.pinvh(KB)

    # u_test is the mean of the data
    A_invKb = A.dot(invKB)

    u_test = np.tensordot(A_invKb, u_map.T, axes=1).T
    # equivalent to:
    # nu, p, _ = u_map.shape
    # u_test = np.zeros((nu, p, t_test.size))
    # for i in range(nu):
    #     for j in range(p):
    #         u_test[i, j, :] = A_invKb.dot(u_map[i, j, :])

    # Covariance of test data is
    # I_p - AK^{-1}A^T
    test_size = t_test.size
    # test_covariance = np.eye(test_size) - A_invKb.dot(A.T)

    return GWP_construct(u_test, L_map)


def kernel(X, Y=None, var=None, inverse_width=None, normalised=False):
    gamma = 0.5 * inverse_width
    k = var * rbf_kernel(X, Y=Y, gamma=gamma)
    if normalised:
        k *= np.sqrt(inverse_width / (2 * np.pi))
    return k


def periodic_kernel(X, Y=None, inverse_width=None):
    k = kernels.ExpSineSquared(X, Y=Y, length_scale=inverse_width)
    return k


class WishartProcess(TimeGraphLasso):
    def __init__(
            self, theta=100, var_prop=1, mu_prior=1, var_prior=10,
            var_Lprop=10, mu_Lprior=1, var_Lprior=1, n_iter=500, burn_in=None,
            verbose=False, assume_centered=False, kernel=None):
        self.n_iter = n_iter
        self.burn_in = n_iter // 4 if burn_in is None else burn_in
        self.verbose = verbose
        self.assume_centered = assume_centered

        # parameter initialisation
        self.theta = theta  # Inverse width
        self.var_prop = var_prop
        self.mu_prior = mu_prior
        self.var_prior = var_prior

        self.var_Lprop = var_Lprop
        self.mu_Lprior = mu_Lprior
        self.var_Lprior = var_Lprior
        self.kernel = kernel

    def fit(self, X, y):
        """Fit the WishartProcess model to X.

        Parameters
        ----------
        X : ndarray, shape = (n_samples * n_times, n_dimensions)
            Data matrix.
        y : ndarray, shape = (n_times,)
            Indicate the temporal belonging of each matrix.
        """
        # Covariance does not make sense for a single feature
        X, y = check_X_y(
            X, y, accept_sparse=False, dtype=np.float64, order="C",
            ensure_min_features=2, estimator=self)

        n_dimensions = X.shape[1]
        self.classes_, n_samples = np.unique(y, return_counts=True)
        n_times = self.classes_.size

        # n_samples = np.array([x.shape[0] for x in X])
        if self.assume_centered:
            self.location_ = np.zeros((n_times, n_dimensions))
        else:
            self.location_ = np.array(
                [X[y == cl].mean(0) for cl in self.classes_])

        # X = (X - self.location_).transpose(1, 2, 0)  # put time last
        X_center = [
            X[y == cl] - self.location_[i]
            for i, cl in enumerate(self.classes_)
        ]
        if self.kernel is None or self.kernel.lower() == 'rbf':
            kern = partial(kernel, var=1)
        else:
            kern = periodic_kernel

        self.likelihood = partial(stats.t_mvn_logpdf, X_center)

        self.nu_ = 1
        samples_u, loglikes, lps, Ls = fit(
            self.theta, self.var_prop, self.mu_prior, self.var_prior,
            self.var_Lprop, self.mu_Lprior, self.var_Lprior, kern=kern,
            t=self.classes_, nu=self.nu_, p=n_dimensions, n_iter=self.n_iter,
            verbose=self.verbose, likelihood=self.likelihood)

        # Burn in
        self.lps_after_burnin = lps[self.burn_in:]
        self.samples_u_after_burnin = samples_u[self.burn_in:]
        self.loglikes_after_burnin = loglikes[self.burn_in:]
        self.Ls_after_burnin = Ls[self.burn_in:]

        # % Select the best hyperparameters based on the loglikes_after_burnin
        pos = np.argmax(self.loglikes_after_burnin)
        self.lmap = self.lps_after_burnin[pos]
        self.Lmap = self.Ls_after_burnin[pos]
        self.u_map = self.samples_u_after_burnin[pos]
        self.D_map = GWP_construct(self.u_map, self.Lmap)

        # compatibility with sklearn
        self.covariance_ = self.D_map.T
        self.precision_ = np.array(
            [linalg.pinvh(cov) for cov in self.covariance_])
        return self

    def score(self, X, y):
        """Computes the log-likelihood of a Gaussian data set with
        `self.covariance_` as an estimator of its covariance matrix.

        Parameters
        ----------
        X : array-like, shape = [n_samples * n_times, n_features]
            Test data of which we compute the likelihood, where n_samples is
            the number of samples and n_features is the number of features.
            X is assumed to be drawn from the same distribution than
            the data used in fit (including centering).

        y : not used, present for API consistence purpose.

        Returns
        -------
        res : float
            The likelihood of the data set with `self.covariance_` as an
            estimator of its covariance matrix.

        """
        # Covariance does not make sense for a single feature
        X, y = check_X_y(
            X, y, accept_sparse=False, dtype=np.float64, order="C",
            ensure_min_features=2, estimator=self)

        X_center = np.array(
            [
                X[y == cl] - self.location_[i]
                for i, cl in enumerate(self.classes_)
            ])
        logp = stats.t_mvn_logpdf(X_center, self.D_map)
        return logp



"""Update rules."""
from __future__ import division


def update_rho(rho, rnorm, snorm, iteration=None, mu=10, tau_inc=2, tau_dec=2):
    """See Boyd pag 20-21 for details.

    Parameters
    ----------
    rho : float
    """
    if rnorm > mu * snorm:
        return tau_inc * rho
    elif snorm > mu * rnorm:
        return rho / tau_dec
    return rho


def update_gamma(gamma, iteration, eps=1e-4):
    """Update `gamma` for forward-backward splitting."""
    if iteration % 20 == 0:
        gamma /= 2.
    return max(gamma, eps)



"""Plotting utils."""
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Ellipse
from scipy import linalg


def plot_cov_2d(means, cov, sdwidth=1.0, npts=50, ax=None, c=None):
    """Plot a 2D covariance matrix.

    Parameters
    ----------
    means : list
        List of two means for the two dimensions of the covariance.
    cov : ndarray
        Covariance matrix (must be 2-dimensional).
    sdwidth : float, optional
        Width of the standard deviation.
    npts : int, optional
        Number of points used to generate the plot.
    ax : plt.ax, optional
        Pass an axis to plot the covariance there,
    c : str, optional
        Color of the plot.
    """
    tt = np.linspace(0, 2 * np.pi, 100)
    ap = np.array([np.cos(tt), np.sin(tt)])

    d, v = linalg.eigh(cov)
    d = sdwidth * np.sqrt(d)

    bp = (v.dot(np.diag(d)).dot(ap)) + means.ravel()[:, None]
    plot = plt.plot if ax is None else ax.plot
    plot(bp[0], bp[1], c=c)
    return cov[0, 1] # return their correlation


def plot_cov_ellipse(pos, cov, nstd=2, ax=None, **kwargs):
    """Plots an `nstd` sigma error ellipse based on the specified covariance.

    Additional keyword arguments are passed on to the ellipse patch artist.

    Parameters
    ----------
    cov : The 2x2 covariance matrix to base the ellipse on
    pos : The location of the center of the ellipse. Expects a 2-element
        sequence of [x0, y0].
    nstd : The radius of the ellipse in numbers of standard deviations.
        Defaults to 2 standard deviations.
    ax : The axis that the ellipse will be plotted on. Defaults to the
        current axis.
    Additional keyword arguments are pass on to the ellipse patch.

    Returns
    -------
    A matplotlib ellipse artist
    """

    def eigsorted(cov):
        vals, vecs = np.linalg.eigh(cov)
        order = vals.argsort()[::-1]
        return vals[order], vecs[:, order]

    vals, vecs = eigsorted(cov)
    theta = np.degrees(np.arctan2(*vecs[:, 0][::-1]))

    # Width and height are "full" widths, not radius
    width, height = 2 * nstd * np.sqrt(vals)

    if width == 0:
        width += 0.1
    if height == 0:
        height += 0.1

    if ax is None:
        ax = plt.gca()
        plt.xlim([-(width + pos[0] + 0.1), width + pos[0] + 0.1])
        plt.ylim([-(height + pos[1] + 0.1), height + pos[1] + 0.1])
    ellip = Ellipse(xy=pos, width=width, height=height, angle=theta, **kwargs)

    ax.add_artist(ellip)
    return ellip



"""Wrapper for PASPAL (Matlab implementation)."""
import matlab
import matlab.engine
import numpy as np
import os


matlab_engine = None  # matlab.engine.start_matlab()


def group_lasso_overlap_paspal(X, y, groups=(), lamda=0.1, verbose=False,
                               **kwargs):
    """Group Lasso with Overlap via PASPAL (Matlab implementation).

    Parameters
    ----------
    X : ndarray
        Data.
    y : ndarray
        Classes.
    groups : list-type
        Groups of variables.
    lamda : float
        Regularization parameter.
    verbose : boolean
        If True, print debug information.

    Returns
    -------
    coef_
        Coefficient of the Lasso algorithm for each feature.

    """
    global matlab_engine
    if matlab_engine is None or not matlab_engine._check_matlab():
        if verbose:
            print("Starting matlab engine ...")
        # close_engine = True
        matlab_engine = matlab.engine.start_matlab()
    # else:
    #     close_engine = False

    matlab_engine.addpath(
        os.path.join(os.path.abspath(os.path.dirname(__file__)),
                     'matlab/GLO_PRIMAL_DUAL_TOOLBOX/'), nargout=0)

    if verbose:
        print("Start GLOPRIDU algorithm ...")
    coef_ = matlab_engine.glopridu_algorithm(
        matlab.double(X.tolist()),
        matlab.double(y[:, None].tolist()),
        [matlab.int32((np.array(x) + 1).tolist()) for x in groups],  # +1 because of the change of indices
        float(lamda))

    # if close_engine:
    #     matlab_engine.quit()
    coef_ = np.asarray(coef_).ravel()
    return coef_, None, np.nan


def test():
    """Test function for the module."""
    from sklearn.datasets import make_regression
    X, y, coef = make_regression(n_features=10, coef=True, n_informative=5)

    group_lasso_overlap_paspal(X, y, np.ones(10), 0.1)



import matlab
import matlab.engine
import numpy as np
import os

matlab_engine = matlab.engine.start_matlab()


def total_variation_condat(y, lamda, verbose=False):
    # if matlab_engine is None or not matlab_engine._check_matlab():
    #     if verbose:
    #         print("Starting matlab engine ...")
    #     close_engine = True
    #     matlab_engine = matlab.engine.start_matlab()
    # else:
    #     close_engine = False
    global matlab_engine

    matlab_engine.addpath(
        os.path.abspath(os.path.dirname(__file__)), nargout=0)

    if verbose:
        print("Start GLOPRIDU algorithm ...")
    x = matlab_engine.TV_Condat_v2(
        matlab.double(y[:, None].tolist()), float(lamda))

    # if close_engine:
    #     matlab_engine.quit()
    x = np.asarray(x).ravel()
    return x



"""docstring."""
import numpy as np
from functools import partial
from sklearn.base import is_classifier, clone
from sklearn.metrics.scorer import _check_multimetric_scoring
from sklearn.model_selection import cross_val_score
from sklearn.model_selection._split import check_cv
from sklearn.model_selection._search import BaseSearchCV

import GPyOpt
from GPyOpt.core.task.objective import SingleObjective
from GPyOpt.core.task.cost import CostModel
from GPyOpt.core.task.space import Design_space
from GPyOpt.util.arguments_manager import ArgumentsManager
from GPyOpt.optimization.acquisition_optimizer import AcquisitionOptimizer


class BayesianOptimization(GPyOpt.methods.BayesianOptimization, BaseSearchCV):
    """Wrapper for BayesianOptimization object from GPyOpt package.

    This is done to use estimator in the same way as scikit-learn.
    For more information, see GPyOpt package.
    """

    def __init__(self, estimator, domain=None, constraints=None,
                 cost_withGradients=None, model_type='GP', X=None, Y=None,
                 initial_design_numdata=5, initial_design_type='random',
                 acquisition_type='EI', normalize_Y=True, exact_feval=False,
                 acquisition_optimizer_type='lbfgs', model_update_interval=1,
                 evaluator_type='sequential', batch_size=1, num_cores=1,
                 verbosity_model=False, verbosity=False, de_duplication=False,
                 max_iter=50, refit=True, cv=None, scoring=None, n_jobs=1,
                 verbose=False,
                 **kwargs):
        """Initialise the estimator."""
        # super(BayesianOptimization, self).__init__(
        #     f=f, domain=domain, constraints=constraints,
        #     cost_withGradients=cost_withGradients, model_type=model_type,
        #     X=X, Y=Y, initial_design_numdata=initial_design_numdata,
        #     initial_design_type=initial_design_type,
        #     acquisition_type=acquisition_type, normalize_Y=normalize_Y,
        #     exact_feval=exact_feval,
        #     acquisition_optimizer_type=acquisition_optimizer_type,
        #     model_update_interval=model_update_interval,
        #     evaluator_type=evaluator_type, batch_size=batch_size,
        #     num_cores=num_cores,
        #     verbosity=verbosity, verbosity_model=verbosity_model,
        #     maximize=True, de_duplication=de_duplication, **kwargs)
        self.modular_optimization = False
        self.initial_iter = True
        self.verbosity = verbosity
        self.verbosity_model = verbosity_model
        self.model_update_interval = model_update_interval
        self.de_duplication = de_duplication
        self.kwargs = kwargs

        # --- Handle the arguments passed via kargs
        self.problem_config = ArgumentsManager(kwargs)

        # --- CHOOSE design space
        self.constraints = constraints
        self.domain = domain or {}
        self.space = Design_space(self.domain, self.constraints)

        # --- CHOOSE objective function
        self.objective_name = kwargs.get('objective_name', '') or 'no_name'
        self.batch_size = batch_size
        self.num_cores = num_cores
        # self.maximize = True
        # self.objective = None

        # --- CHOOSE the cost model
        self.cost = CostModel(cost_withGradients)

        # --- CHOOSE initial design
        self.X = X  # parameters
        self.Y = Y  # evaluation
        self.initial_design_type = initial_design_type
        self.initial_design_numdata = initial_design_numdata

        # --- CHOOSE the model type. If an instance of a GPyOpt model is passed (possibly user defined), it is used.
        # note that this 2 options are not used with the predefined model
        self.model_type = model_type
        self.exact_feval = exact_feval
        self.normalize_Y = normalize_Y

        if 'model' in self.kwargs and isinstance(
                kwargs['model'], GPyOpt.models.base.BOModel):
            self.model = kwargs['model']
            self.model_type = 'User defined model used.'
            if self.verbose:
                print('Using a model defined by the used.')
        else:
            self.model = self._model_chooser()

        # --- CHOOSE the acquisition optimizer_type
        # This states how the discrete variables are handled (exact search or rounding)
        self.acquisition_optimizer_type = acquisition_optimizer_type
        self.acquisition_optimizer = AcquisitionOptimizer(
            self.space, self.acquisition_optimizer_type, model=self.model)

        # --- CHOOSE acquisition function. If an instance of an acquisition is passed (possibly user defined), it is used.
        self.acquisition_type = acquisition_type
        if 'acquisition' in self.kwargs and isinstance(
                kwargs['acquisition'], GPyOpt.acquisitions.AcquisitionBase):
            self.acquisition = kwargs['acquisition']
            self.acquisition_type = 'User defined acquisition used.'
            if self.verbose:
                print('Using an acquisition defined by the used.')
        else:
            self.acquisition = self._acquisition_chooser()

        # --- CHOOSE evaluator method
        self.evaluator_type = evaluator_type
        self.evaluator = self._evaluator_chooser()

        # --- Create optimization space
        self.cost = CostModel(self.cost)
        self.normalization_type = 'stats'  # not added in the API

        self.estimator = estimator
        self.cv = cv
        self.max_iter = max_iter
        self.refit = refit
        self.scoring = scoring
        self.verbose = verbose
        self.n_jobs = n_jobs

    @property
    def param_names(self):
        return [element['name'] for element in self.domain]

    def fit(self, X, y=None, groups=None, **fit_params):
        """Run fit with all sets of parameters.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples] or [n_samples, n_output], optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of the estimator

        """
        cv = check_cv(self.cv, y, classifier=is_classifier(self.estimator))

        scorers, self.multimetric_ = _check_multimetric_scoring(
            self.estimator, scoring=self.scoring)

        score_function = partial(
            cross_val_score, X=X, y=y, groups=groups, scoring=self.scoring,
            cv=cv, n_jobs=self.n_jobs, verbose=self.verbose,
            fit_params=fit_params)
        self.f = partial(
            _fit_score, mdl=self.estimator, param_names=self.param_names,
            score_function=score_function)

        self.objective = SingleObjective(
            self.f, self.batch_size, self.objective_name)
        self._init_design_chooser()

        self.run_optimization(max_iter=self.max_iter, verbosity=self.verbosity)

        self.best_index_ = self.Y.argmin()
        self.best_params_ = dict(zip(self.param_names,
                                     10 ** self.X[self.best_index_]))
        self.best_score_ = self.Y[self.Y.argmin()]

        # Store the only scorer not as a dict for single metric evaluation
        self.scorer_ = scorers if self.multimetric_ else scorers['score']

        if self.refit:
            self.best_estimator_ = clone(self.estimator).set_params(
                **self.best_params_)
            if y is not None:
                self.best_estimator_.fit(X, y, **fit_params)
            else:
                self.best_estimator_.fit(X, **fit_params)

        return self

    def plot_acquisition(self):
        old_X, old_Y = self.X, self.Y
        # remove inf, to avoid error in plot_acquisition
        self.X = self.X[~np.isinf(self.Y).ravel()]
        self.Y = self.Y[~np.isinf(self.Y).ravel()]
        super(BayesianOptimization, self).plot_acquisition()
        # restore original X and Y
        self.X, self.Y = old_X, old_Y


def _fit_score(x, score_function, mdl=None, param_names=None):
    x = 10 ** np.atleast_2d(x)
    return - np.array([np.mean(score_function(mdl.set_params(
        **dict(zip(param_names, pars)))))
                       for pars in x])[:, None]



"""Sparse inverse covariance selection over time via ADMM.

More information can be found in the paper linked at:
https://arxiv.org/abs/1703.01958
"""
from __future__ import division

import warnings

import numpy as np
import scipy.sparse as sp
from scipy import linalg
from six.moves import map, range, zip
from sklearn.covariance import empirical_covariance, log_likelihood
from sklearn.utils.extmath import squared_norm
from sklearn.utils.validation import check_array

from regain.covariance.graph_lasso_ import GraphLasso, logl
from regain.norm import l1_od_norm
from regain.prox import prox_logdet, soft_thresholding
from regain.update_rules import update_rho
from regain.utils import convergence, error_norm_time
from regain.validation import check_array_dimensions, check_norm_prox


def loss(S, K, n_samples=None):
    """Loss function for time-varying graphical lasso."""
    if n_samples is None:
        n_samples = np.ones(S.shape[0])
    return sum(-ni * logl(emp_cov, precision)
               for emp_cov, precision, ni in zip(S, K, n_samples))


def objective(n_samples, S, K, Z_0, Z_1, Z_2, alpha, beta, psi):
    """Objective function for time-varying graphical lasso."""
    obj = loss(S, K, n_samples=n_samples)

    if isinstance(alpha, np.ndarray):
        obj += sum(l1_od_norm(a * z) for a, z in zip(alpha, Z_0))
    else:
        obj += alpha * sum(map(l1_od_norm, Z_0))
    
    if isinstance(beta, np.ndarray):
        obj += sum(b[0][0] * m for b, m in zip(beta, map(psi, Z_2 - Z_1)))
    else:
        obj += beta * sum(map(psi, Z_2 - Z_1))
    return obj


def time_graph_lasso(
        emp_cov, alpha=0.01, rho=1, beta=1, max_iter=100, n_samples=None,
        verbose=False, psi='laplacian', tol=1e-4, rtol=1e-4,
        return_history=False, return_n_iter=True, mode='admm',
        update_rho_options=None, compute_objective=True,
        stop_at=None, stop_when=1e-4):
    """Time-varying graphical lasso solver.

    Solves the following problem via ADMM:
        minimize  trace(S*X) - log det X + lambda*||X||_1
            + beta sum_{i=2}^T Psi(K_i - K_{i-1})

    where S is the empirical covariance of the data
    matrix D (training observations by features).

    Parameters
    ----------
    emp_cov : ndarray, shape (n_features, n_features)
        Empirical covariance of data.
    alpha, beta : float, optional
        Regularisation parameter.
    rho : float, optional
        Augmented Lagrangian parameter.
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Absolute tolerance for convergence.
    rtol : float, optional
        Relative tolerance for convergence.
    return_history : bool, optional
        Return the history of computed values.

    Returns
    -------
    X : numpy.array, 2-dimensional
        Solution to the problem.
    history : list
        If return_history, then also a structure that contains the
        objective value, the primal and dual residual norms, and tolerances
        for the primal and dual residual norms at each iteration.

    """
    psi, prox_psi, psi_node_penalty = check_norm_prox(psi)

    n_times, _, n_features = emp_cov.shape
    covariance_ = emp_cov.copy()
    covariance_ *= 0.95
    K = np.empty_like(emp_cov)
    for i, (c, e) in enumerate(zip(covariance_, emp_cov)):
        c.flat[::n_features + 1] = e.flat[::n_features + 1]
        K[i] = linalg.pinvh(c)

    # K = np.zeros_like(emp_cov)
    Z_0 = K.copy() # np.zeros_like(emp_cov)
    Z_1 = K.copy()[:-1] # np.zeros_like(emp_cov)[:-1]
    Z_2 = K.copy()[1:] # np.zeros_like(emp_cov)[1:]
    # Z_0 = K.copy()
    # Z_1 = K.copy()[:-1]
    # Z_2 = K.copy()[1:]

    U_0 = np.zeros_like(Z_0)
    U_1 = np.zeros_like(Z_1)
    U_2 = np.zeros_like(Z_2)

    Z_0_old = np.zeros_like(Z_0)
    Z_1_old = np.zeros_like(Z_1)
    Z_2_old = np.zeros_like(Z_2)

    # divisor for consensus variables, accounting for two less matrices
    divisor = np.full(emp_cov.shape[0], 3, dtype=float)
    divisor[0] -= 1
    divisor[-1] -= 1

    if n_samples is None:
        n_samples = np.ones(emp_cov.shape[0])

    checks = [convergence(obj=objective(
        n_samples, emp_cov, Z_0, K, Z_1, Z_2, alpha, beta, psi))]
    for iteration_ in range(max_iter):
        # update K
        A = Z_0 - U_0
        A[:-1] += Z_1 - U_1
        A[1:] += Z_2 - U_2
        A /= divisor[:, None, None]
        # soft_thresholding_ = partial(soft_thresholding, lamda=alpha / rho)
        # K = np.array(map(soft_thresholding_, A))
        A += A.transpose(0, 2, 1)
        A /= 2.

        A *= - rho * divisor[:, None, None] / n_samples[:, None, None]
        A += emp_cov

        K = np.array([prox_logdet(a, lamda=ni / (rho * div))
                      for a, div, ni in zip(A, divisor, n_samples)])

        # update Z_0
        A = K + U_0
        A += A.transpose(0, 2, 1)
        A /= 2.
        Z_0 = soft_thresholding(A, lamda=alpha / rho)

        # other Zs
        A_1 = K[:-1] + U_1
        A_2 = K[1:] + U_2
        if not psi_node_penalty:
            prox_e = prox_psi(A_2 - A_1, lamda=2. * beta / rho)
            Z_1 = .5 * (A_1 + A_2 - prox_e)
            Z_2 = .5 * (A_1 + A_2 + prox_e)
        else:
            Z_1, Z_2 = prox_psi(np.concatenate((A_1, A_2), axis=1),
                                lamda=.5 * beta / rho,
                                rho=rho, tol=tol, rtol=rtol, max_iter=max_iter)

        # update residuals
        U_0 += K - Z_0
        U_1 += K[:-1] - Z_1
        U_2 += K[1:] - Z_2

        # diagnostics, reporting, termination checks
        rnorm = np.sqrt(squared_norm(K - Z_0) +
                        squared_norm(K[:-1] - Z_1) +
                        squared_norm(K[1:] - Z_2))

        snorm = rho * np.sqrt(squared_norm(Z_0 - Z_0_old) +
                              squared_norm(Z_1 - Z_1_old) +
                              squared_norm(Z_2 - Z_2_old))

        obj = objective(
            n_samples, emp_cov, Z_0, K, Z_1, Z_2, alpha, beta, psi) \
            if compute_objective else np.nan

        check = convergence(
            obj=obj, rnorm=rnorm, snorm=snorm,
            e_pri=np.sqrt(K.size + 2 * Z_1.size) * tol + rtol * max(
                np.sqrt(squared_norm(Z_0) + squared_norm(Z_1) + squared_norm(Z_2)),
                np.sqrt(squared_norm(K) + squared_norm(K[:-1]) + squared_norm(K[1:]))),
            e_dual=np.sqrt(K.size + 2 * Z_1.size) * tol + rtol * rho * np.sqrt(
                squared_norm(U_0) + squared_norm(U_1) + squared_norm(U_2)),
            # precision=Z_0.copy()
        )
        Z_0_old = Z_0.copy()
        Z_1_old = Z_1.copy()
        Z_2_old = Z_2.copy()

        if verbose:
            print("obj: %.4f, rnorm: %.4f, snorm: %.4f,"
                  "eps_pri: %.4f, eps_dual: %.4f" % check[:5])

        checks.append(check)
        if stop_at is not None:
            if abs(check.obj - stop_at) / abs(stop_at) < stop_when:
                break

        if check.rnorm <= check.e_pri and check.snorm <= check.e_dual:
            break

        rho_new = update_rho(rho, rnorm, snorm, iteration=iteration_,
                             **(update_rho_options or {}))
        # scaled dual variables should be also rescaled
        U_0 *= rho / rho_new
        U_1 *= rho / rho_new
        U_2 *= rho / rho_new
        rho = rho_new
    else:
        warnings.warn("Objective did not converge.")

    covariance_ = np.array([linalg.pinvh(x) for x in Z_0])
    return_list = [Z_0, covariance_]
    if return_history:
        return_list.append(checks)
    if return_n_iter:
        return_list.append(iteration_ + 1)
    return return_list


class TimeGraphLasso(GraphLasso):
    """Sparse inverse covariance estimation with an l1-penalized estimator.

    Parameters
    ----------
    alpha : positive float, default 0.01
        Regularization parameter for precision matrix. The higher alpha,
        the more regularization, the sparser the inverse covariance.

    beta : positive float, default 1
        Regularization parameter to constrain precision matrices in time.
        The higher beta, the more regularization,
        and consecutive precision matrices in time are more similar.

    psi : {'laplacian', 'l1', 'l2', 'linf', 'node'}, default 'laplacian'
        Type of norm to enforce for consecutive precision matrices in time.

    rho : positive float, default 1
        Augmented Lagrangian parameter.

    over_relax : positive float, deafult 1
        Over-relaxation parameter (typically between 1.0 and 1.8).

    tol : positive float, default 1e-4
        Absolute tolerance to declare convergence.

    rtol : positive float, default 1e-4
        Relative tolerance to declare convergence.

    max_iter : integer, default 100
        The maximum number of iterations.

    verbose : boolean, default False
        If verbose is True, the objective function, rnorm and snorm are
        printed at each iteration.

    assume_centered : boolean, default False
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    time_on_axis : {'first', 'last'}, default 'first'
        If data have time as the last dimension, set this to 'last'.
        Useful to use scikit-learn functions as train_test_split.

    update_rho_options : dict, default None
        Options for the update of rho. See `update_rho` function for details.

    compute_objective : boolean, default True
        Choose if compute the objective function during iterations
        (only useful if `verbose=True`).

    mode : {'admm'}, default 'admm'
        Minimisation algorithm. At the moment, only 'admm' is available,
        so this is ignored.

    Attributes
    ----------
    covariance_ : array-like, shape (n_times, n_features, n_features)
        Estimated covariance matrix

    precision_ : array-like, shape (n_times, n_features, n_features)
        Estimated pseudo inverse matrix.

    n_iter_ : int
        Number of iterations run.

    """

    def __init__(
            self, alpha=0.01, beta=1., mode='admm', rho=1.,
            time_on_axis='first', tol=1e-4, rtol=1e-4, psi='laplacian',
            max_iter=100, verbose=False, assume_centered=False,
            return_history=False, update_rho_options=None,
            compute_objective=True, stop_at=None, stop_when=1e-4,
            suppress_warn_list=False):
        super(TimeGraphLasso, self).__init__(
            alpha=alpha, rho=rho, tol=tol, rtol=rtol, max_iter=max_iter,
            verbose=verbose, assume_centered=assume_centered, mode=mode,
            update_rho_options=update_rho_options,
            compute_objective=compute_objective)
        self.beta = beta
        self.psi = psi
        self.time_on_axis = time_on_axis
        self.return_history = return_history
        self.stop_at = stop_at
        self.stop_when = stop_when
        self.suppress_warn_list = suppress_warn_list

    def get_observed_precision(self):
        """Getter for the observed precision matrix.

        Returns
        -------
        precision_ : array-like,
            The precision matrix associated to the current covariance object.

        """
        return self.get_precision()

    def _fit(self, emp_cov, n_samples):
        """Fit the TimeGraphLasso model to X.

        Parameters
        ----------
        emp_cov : ndarray, shape (n_time, n_features, n_features)
            Empirical covariance of data.

        """
        out = time_graph_lasso(
                emp_cov, alpha=self.alpha, rho=self.rho,
                beta=self.beta, mode=self.mode, n_samples=n_samples,
                tol=self.tol, rtol=self.rtol, psi=self.psi,
                max_iter=self.max_iter, verbose=self.verbose,
                return_n_iter=True, return_history=self.return_history,
                update_rho_options=self.update_rho_options,
                compute_objective=self.compute_objective,
                stop_at=self.stop_at, stop_when=self.stop_when)
        if self.return_history:
            self.precision_, self.covariance_, self.history_, self.n_iter_ = out
        else:
            self.precision_, self.covariance_, self.n_iter_ = out
        return self

    def fit(self, X, y=None):
        """Fit the TimeGraphLasso model to X.

        Parameters
        ----------
        X : ndarray, shape (n_time, n_samples, n_features), or
                (n_samples, n_features, n_time)
            Data from which to compute the covariance estimate.
            If shape is (n_samples, n_features, n_time), then set
            `time_on_axis = 'last'`.
        y : (ignored)

        """
        if sp.issparse(X):
            raise TypeError("sparse matrices not supported.")

        X = check_array_dimensions(
            X, n_dimensions=3, time_on_axis=self.time_on_axis,
            suppress_warn_list=self.suppress_warn_list)

        is_list = isinstance(X, list)
        # Covariance does not make sense for a single feature
        X = np.array([check_array(x, ensure_min_features=2,
                      ensure_min_samples=2, estimator=self) for x in X])
        if is_list:
            n_times = len(X)
            if np.unique([x.shape[1] for x in X]).size != 1:
                raise ValueError("Input data cannot have different number "
                                 "of variables.")
            n_dimensions = X[0].shape[1]
        else:
            n_times = X.shape[0]
            n_dimensions = X.shape[2]

        n_samples = np.array([x.shape[0] for x in X])

        if self.assume_centered:
            self.location_ = np.zeros((n_times, 1, n_dimensions))
        else:
            mean = np.array([x.mean(0) for x in X]) if is_list else X.mean(1)
            self.location_ = mean.reshape(n_times, 1, n_dimensions)
        emp_cov = np.array([empirical_covariance(
            x, assume_centered=self.assume_centered) for x in X])

        return self._fit(emp_cov, n_samples)

    def score(self, X_test, y=None):
        """Computes the log-likelihood of a Gaussian data set with
        `self.covariance_` as an estimator of its covariance matrix.

        Parameters
        ----------
        X_test : array-like, shape = [n_samples, n_features]
            Test data of which we compute the likelihood, where n_samples is
            the number of samples and n_features is the number of features.
            X_test is assumed to be drawn from the same distribution than
            the data used in fit (including centering).

        y : not used, present for API consistence purpose.

        Returns
        -------
        res : float
            The likelihood of the data set with `self.covariance_` as an
            estimator of its covariance matrix.

        """
        if sp.issparse(X_test):
            raise TypeError("sparse matrices not supported.")

        X_test = check_array_dimensions(
            X_test, n_dimensions=3, time_on_axis=self.time_on_axis)

        # Covariance does not make sense for a single feature
        X_test = np.array([
            check_array(x, ensure_min_features=2,
                        ensure_min_samples=2, estimator=self) for x in X_test])

        # compute empirical covariance of the test set
        test_cov = np.array([empirical_covariance(
            x, assume_centered=True) for x in X_test - self.location_])

        n_samples = np.array([x.shape[0] for x in X_test])
        res = sum(n * log_likelihood(S, K) for S, K, n in zip(
            test_cov, self.get_observed_precision(), n_samples))

        # ALLA  MATLAB1
        # ranks = [np.linalg.matrix_rank(L) for L in self.latent_]
        # scores_ranks = np.square(ranks-np.sqrt(L.shape[1]))

        return res  # - np.sum(scores_ranks)

    def error_norm(self, comp_cov, norm='frobenius', scaling=True,
                   squared=True):
        """Compute the Mean Squared Error between two covariance estimators.
        (In the sense of the Frobenius norm).

        Parameters
        ----------
        comp_cov : array-like, shape = [n_features, n_features]
            The covariance to compare with.

        norm : str
            The type of norm used to compute the error. Available error types:
            - 'frobenius' (default): sqrt(tr(A^t.A))
            - 'spectral': sqrt(max(eigenvalues(A^t.A))
            where A is the error ``(comp_cov - self.covariance_)``.

        scaling : bool
            If True (default), the squared error norm is divided by n_features.
            If False, the squared error norm is not rescaled.

        squared : bool
            Whether to compute the squared error norm or the error norm.
            If True (default), the squared error norm is returned.
            If False, the error norm is returned.

        Returns
        -------
        The Mean Squared Error (in the sense of the Frobenius norm) between
        `self` and `comp_cov` covariance estimators.

        """
        return error_norm_time(self.covariance_, comp_cov, norm=norm,
                               scaling=scaling, squared=squared)

    def mahalanobis(self, observations):
        """Computes the squared Mahalanobis distances of given observations.

        Parameters
        ----------
        observations : array-like, shape = [n_observations, n_features]
            The observations, the Mahalanobis distances of the which we
            compute. Observations are assumed to be drawn from the same
            distribution than the data used in fit.

        Returns
        -------
        mahalanobis_distance : array, shape = [n_observations,]
            Squared Mahalanobis distances of the observations.

        """
        if self.time_on_axis == 'last':
            # put time as first dimension
            observations = observations.transpose(2, 0, 1)
        precision = self.get_observed_precision()
        # compute mahalanobis distances
        sum_ = 0.
        for obs, loc in zip(observations, self.location_):
            centered_obs = observations - self.location_
            sum_ += np.sum(
                np.dot(centered_obs, precision) * centered_obs, 1)

        mahalanobis_dist = sum_ / len(observations)
        return mahalanobis_dist



"""Discriminant analysis with REGAIN.

XXX This module is work in progress.
TODO Remove deprecated `decision_function2`
"""

import numpy as np
from scipy import linalg
from six.moves import xrange
from sklearn.base import BaseEstimator
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.utils import check_array, check_X_y, deprecated
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_is_fitted

from regain.covariance.graph_lasso_ import fast_logdet
from regain.utils import ensure_posdef

__all__ = ("DiscriminantAnalysis", "PrecomputedDiscriminantAnalysis")


class DiscriminantAnalysis(QuadraticDiscriminantAnalysis):
    """Quadratic Discriminant Analysis using LTGL

    A classifier with a quadratic decision boundary, generated
    by fitting class conditional densities to the data
    and using Bayes' rule.

    The model fits a Gaussian density to each class using LTGL.

    Parameters
    ----------
    estimator : class
        Estimator to compute precision and covariance matrices

    priors : array, optional, shape = [n_classes]
        Priors on classes

    Attributes
    ----------
    covariance_ : list of array-like, shape = [n_features, n_features]
        Covariance matrices of each class.

    means_ : array-like, shape = [n_classes, n_features]
        Class means.

    priors_ : array-like, shape = [n_classes]
        Class priors (sum to 1).

    rotations_ : list of arrays
        For each class k an array of shape [n_features, n_k], with
        ``n_k = min(n_features, number of elements in class k)``
        It is the rotation of the Gaussian distribution, i.e. its
        principal axis.

    scalings_ : list of arrays
        For each class k an array of shape [n_k]. It contains the scaling
        of the Gaussian distributions along its principal axes, i.e. the
        variance in the rotated coordinate system.
    """

    def __init__(self, estimator, ensure_posdef=False, priors=None):
        self.estimator = estimator
        self.priors = np.asarray(priors) if priors is not None else None
        self.ensure_posdef = ensure_posdef

    def fit(self, X, y):
        """Fit the model according to the given training data and parameters.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array, shape = [n_samples]
            Target values (integers)
        """
        if not hasattr(self.estimator, "fit"):
            raise ValueError("estimator does not implement a `fit` method.")
        X, y = check_X_y(X, y)
        check_classification_targets(y)
        self.classes_, y = np.unique(y, return_inverse=True)
        n_samples, n_features = X.shape
        n_classes = len(self.classes_)
        if n_classes < 2:
            raise ValueError(
                'The number of classes has to be greater than'
                ' one; got %d class' % (n_classes))
        if self.priors is None:
            self.priors_ = np.bincount(y) / float(n_samples)
        else:
            self.priors_ = self.priors
        means = []
        data = []
        for ind in xrange(n_classes):
            Xg = X[y == ind, :]
            meang = Xg.mean(0)
            means.append(meang)
            if len(Xg) == 1:
                raise ValueError(
                    'y has only 1 sample in class %s, covariance '
                    'is ill defined.' % str(self.classes_[ind]))
            data.append(Xg)

        self.estimator.fit(data, y)
        self.precision_ = self.estimator.precision_

        if hasattr(self.estimator, "covariance_"):
            self.covariance_ = self.estimator.covariance_
        elif not self.ensure_posdef:
            self.covariance_ = np.array(
                [linalg.pinvh(p) for p in self.precision_])

        if self.ensure_posdef:
            # replace diagonal
            ensure_posdef(self.precision_, inplace=True)
            self.covariance_ = np.array(
                [linalg.pinvh(p) for p in self.precision_])
        self.means_ = np.asarray(means)
        return self

    @deprecated(
        "it will be removed in v0.2.0. Use `_decision_function` instead")
    def _decision_function2(self, X):
        check_is_fitted(self, 'classes_')

        X = check_array(X)
        precisions = self.get_observed_precision()
        norm2 = []
        for i in range(len(self.classes_)):
            Xm = X - self.means_[i]
            # X2 = np.dot(Xm, R * (S ** (-0.5)))
            X2 = np.linalg.multi_dot((Xm, precisions[i], Xm.T))
            norm2.append(np.diag(X2))
        norm2 = np.array(norm2).T  # shape = [len(X), n_classes]
        u = np.asarray([-fast_logdet(s) for s in precisions])
        return (-0.5 * (norm2 + u) + np.log(self.priors_))

    def _decision_function(self, X):
        rotations, scalings = [], []
        for p in self.covariance_:
            R, S, Rt = np.linalg.svd(p, full_matrices=False)
            rotations.append(Rt.T)
            scalings.append(S)
        self.rotations_ = rotations
        self.scalings_ = scalings
        return super(DiscriminantAnalysis, self)._decision_function(X)

    def score(self, X, y, sample_weight=None):
        """Returns the mean accuracy on the given test data and labels.

        In multi-label classification, this is the subset accuracy
        which is a harsh metric since you require for each sample that
        each label set be correctly predicted.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        y : array-like, shape = (n_samples) or (n_samples, n_outputs)
            True labels for X.

        sample_weight : array-like, shape = [n_samples], optional
            Sample weights.

        Returns
        -------
        score : float
            Mean accuracy of self.predict(X) wrt. y.

        """
        from sklearn.metrics import accuracy_score
        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)


class PrecomputedDiscriminantAnalysis(BaseEstimator):
    def __init__(self, precision):
        self.precision_ = np.array(precision)

    def fit(self, X, y=None):
        """Dummy fit."""
        pass



"""Graphical latent variable models selection over time via ADMM."""
from __future__ import division

import warnings
from functools import partial

import numpy as np
from scipy import linalg
from six.moves import map, range, zip
from sklearn.utils.extmath import squared_norm

from regain.covariance.time_graph_lasso_ import TimeGraphLasso
from regain.covariance.time_graph_lasso_ import objective as obj_tgl
from regain.prox import prox_logdet, prox_trace_indicator, soft_thresholding
from regain.update_rules import update_rho
from regain.utils import convergence
from regain.validation import check_norm_prox


def objective(
        S, n_samples, R, Z_0, Z_1, Z_2, W_0, W_1, W_2, alpha, tau, beta, eta,
        psi, phi):
    """Objective function for latent variable time-varying graphical lasso."""
    # obj = sum(- n * logl(s, r) for s, r, n in zip(S, R, n_samples))
    obj = obj_tgl(n_samples, S, R, Z_0, Z_1, Z_2, alpha, beta, psi)
    obj += tau * sum(map(partial(np.linalg.norm, ord='nuc'), W_0))

    if isinstance(eta, np.ndarray):
        obj += sum(b[0][0] * m for b, m in zip(eta, map(phi, W_2 - W_1)))
    else:
        obj += eta * sum(map(phi, W_2 - W_1))
    return obj


def latent_time_graph_lasso(
        emp_cov, alpha=0.01, tau=1., rho=1., beta=1., eta=1., max_iter=100,
        verbose=False, psi='laplacian', phi='laplacian', mode='admm', tol=1e-4,
        rtol=1e-4, assume_centered=False, n_samples=None, return_history=False,
        return_n_iter=True, update_rho_options=None, compute_objective=True):
    r"""Time-varying latent variable graphical lasso solver.

    Solves the following problem via ADMM:
        min sum_{i=1}^T -n_i log_likelihood(K_i-L_i) + alpha ||K_i||_{od,1}
            + tau ||L_i||_*
            + beta sum_{i=2}^T Psi(K_i - K_{i-1})
            + eta sum_{i=2}^T Phi(L_i - L_{i-1})

    where S is the empirical covariance of the data
    matrix D (training observations by features).

    Parameters
    ----------
    emp_cov : ndarray, shape (n_features, n_features)
        Empirical covariance of data.
    alpha, tau, beta, eta : float, optional
        Regularisation parameters.
    rho : float, optional
        Augmented Lagrangian parameter.
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Absolute tolerance for convergence.
    rtol : float, optional
        Relative tolerance for convergence.
    return_history : bool, optional
        Return the history of computed values.

    Returns
    -------
    K, L : numpy.array, 3-dimensional (T x d x d)
        Solution to the problem for each time t=1...T .
    history : list
        If return_history, then also a structure that contains the
        objective value, the primal and dual residual norms, and tolerances
        for the primal and dual residual norms at each iteration.

    """
    psi, prox_psi, psi_node_penalty = check_norm_prox(psi)
    phi, prox_phi, phi_node_penalty = check_norm_prox(phi)

    Z_0 = np.zeros_like(emp_cov)
    Z_1 = np.zeros_like(Z_0)[:-1]
    Z_2 = np.zeros_like(Z_0)[1:]
    W_0 = np.zeros_like(Z_0)
    W_1 = np.zeros_like(Z_1)
    W_2 = np.zeros_like(Z_2)

    X_0 = np.zeros_like(Z_0)
    X_1 = np.zeros_like(Z_1)
    X_2 = np.zeros_like(Z_2)
    U_1 = np.zeros_like(W_1)
    U_2 = np.zeros_like(W_2)

    R_old = np.zeros_like(Z_0)
    Z_1_old = np.zeros_like(Z_1)
    Z_2_old = np.zeros_like(Z_2)
    W_1_old = np.zeros_like(W_1)
    W_2_old = np.zeros_like(W_2)

    # divisor for consensus variables, accounting for two less matrices
    divisor = np.full(emp_cov.shape[0], 3, dtype=float)
    divisor[0] -= 1
    divisor[-1] -= 1

    if n_samples is None:
        n_samples = np.ones(emp_cov.shape[0])

    checks = []
    for iteration_ in range(max_iter):
        # update R
        A = Z_0 - W_0 - X_0
        A += A.transpose(0, 2, 1)
        A /= 2.
        A *= -rho / n_samples[:, None, None]
        A += emp_cov
        # A = emp_cov / rho - A

        R = np.array(
            [prox_logdet(a, lamda=ni / rho) for a, ni in zip(A, n_samples)])

        # update Z_0
        A = R + W_0 + X_0
        A[:-1] += Z_1 - X_1
        A[1:] += Z_2 - X_2
        A /= divisor[:, None, None]
        # soft_thresholding_ = partial(soft_thresholding, lamda=alpha / rho)
        # Z_0 = np.array(map(soft_thresholding_, A))
        Z_0 = soft_thresholding(
            A, lamda=alpha / (rho * divisor[:, None, None]))

        # update Z_1, Z_2
        A_1 = Z_0[:-1] + X_1
        A_2 = Z_0[1:] + X_2
        if not psi_node_penalty:
            prox_e = prox_psi(A_2 - A_1, lamda=2. * beta / rho)
            Z_1 = .5 * (A_1 + A_2 - prox_e)
            Z_2 = .5 * (A_1 + A_2 + prox_e)
        else:
            Z_1, Z_2 = prox_psi(
                np.concatenate((A_1, A_2), axis=1), lamda=.5 * beta / rho,
                rho=rho, tol=tol, rtol=rtol, max_iter=max_iter)

        # update W_0
        A = Z_0 - R - X_0
        A[:-1] += W_1 - U_1
        A[1:] += W_2 - U_2
        A /= divisor[:, None, None]
        A += A.transpose(0, 2, 1)
        A /= 2.

        W_0 = np.array(
            [
                prox_trace_indicator(a, lamda=tau / (rho * div))
                for a, div in zip(A, divisor)
            ])

        # update W_1, W_2
        A_1 = W_0[:-1] + U_1
        A_2 = W_0[1:] + U_2
        if not phi_node_penalty:
            prox_e = prox_phi(A_2 - A_1, lamda=2. * eta / rho)
            W_1 = .5 * (A_1 + A_2 - prox_e)
            W_2 = .5 * (A_1 + A_2 + prox_e)
        else:
            W_1, W_2 = prox_phi(
                np.concatenate((A_1, A_2), axis=1), lamda=.5 * eta / rho,
                rho=rho, tol=tol, rtol=rtol, max_iter=max_iter)

        # update residuals
        X_0 += R - Z_0 + W_0
        X_1 += Z_0[:-1] - Z_1
        X_2 += Z_0[1:] - Z_2
        U_1 += W_0[:-1] - W_1
        U_2 += W_0[1:] - W_2

        # diagnostics, reporting, termination checks
        rnorm = np.sqrt(
            squared_norm(R - Z_0 + W_0) + squared_norm(Z_0[:-1] - Z_1) +
            squared_norm(Z_0[1:] - Z_2) + squared_norm(W_0[:-1] - W_1) +
            squared_norm(W_0[1:] - W_2))

        snorm = rho * np.sqrt(
            squared_norm(R - R_old) + squared_norm(Z_1 - Z_1_old) +
            squared_norm(Z_2 - Z_2_old) + squared_norm(W_1 - W_1_old) +
            squared_norm(W_2 - W_2_old))

        obj = objective(emp_cov, n_samples, R, Z_0, Z_1, Z_2, W_0, W_1, W_2,
                        alpha, tau, beta, eta, psi, phi) \
            if compute_objective else np.nan

        check = convergence(
            obj=obj, rnorm=rnorm, snorm=snorm,
            e_pri=np.sqrt(R.size + 4 * Z_1.size) * tol + rtol * max(
                np.sqrt(
                    squared_norm(R) + squared_norm(Z_1) + squared_norm(Z_2) +
                    squared_norm(W_1) + squared_norm(W_2)),
                np.sqrt(
                    squared_norm(Z_0 - W_0) + squared_norm(Z_0[:-1]) +
                    squared_norm(Z_0[1:]) + squared_norm(W_0[:-1]) +
                    squared_norm(W_0[1:]))),
            e_dual=np.sqrt(R.size + 4 * Z_1.size) * tol + rtol * rho * (
                np.sqrt(
                    squared_norm(X_0) + squared_norm(X_1) + squared_norm(X_2) +
                    squared_norm(U_1) + squared_norm(U_2))))

        R_old = R.copy()
        Z_1_old = Z_1.copy()
        Z_2_old = Z_2.copy()
        W_1_old = W_1.copy()
        W_2_old = W_2.copy()

        if verbose:
            print(
                "obj: %.4f, rnorm: %.4f, snorm: %.4f,"
                "eps_pri: %.4f, eps_dual: %.4f" % check[:5])

        checks.append(check)
        if check.rnorm <= check.e_pri and check.snorm <= check.e_dual:
            break

        rho_new = update_rho(
            rho, rnorm, snorm, iteration=iteration_,
            **(update_rho_options or {}))
        # scaled dual variables should be also rescaled
        X_0 *= rho / rho_new
        X_1 *= rho / rho_new
        X_2 *= rho / rho_new
        U_1 *= rho / rho_new
        U_2 *= rho / rho_new
        rho = rho_new
    else:
        warnings.warn("Objective did not converge.")

    covariance_ = np.array([linalg.pinvh(x) for x in Z_0])
    return_list = [Z_0, W_0, covariance_]
    if return_history:
        return_list.append(checks)
    if return_n_iter:
        return_list.append(iteration_)
    return return_list


class LatentTimeGraphLasso(TimeGraphLasso):
    """Sparse inverse covariance estimation with an l1-penalized estimator.

    Parameters
    ----------
    alpha : positive float, default 0.01
        Regularization parameter for precision matrix. The higher alpha,
        the more regularization, the sparser the inverse covariance.

    tau : positive float, default 1
        Regularization parameter for latent variables matrix. The higher tau,
        the more regularization, the lower rank of the latent matrix.

    beta : positive float, default 1
        Regularization parameter to constrain precision matrices in time.
        The higher beta, the more regularization,
        and consecutive precision matrices in time are more similar.

    psi : {'laplacian', 'l1', 'l2', 'linf', 'node'}, default 'laplacian'
        Type of norm to enforce for consecutive precision matrices in time.

    eta : positive float, default 1
        Regularization parameter to constrain latent matrices in time.
        The higher eta, the more regularization,
        and consecutive latent matrices in time are more similar.

    phi : {'laplacian', 'l1', 'l2', 'linf', 'node'}, default 'laplacian'
        Type of norm to enforce for consecutive latent matrices in time.

    rho : positive float, default 1
        Augmented Lagrangian parameter.

    over_relax : positive float, deafult 1
        Over-relaxation parameter (typically between 1.0 and 1.8).

    tol : positive float, default 1e-4
        Absolute tolerance to declare convergence.

    rtol : positive float, default 1e-4
        Relative tolerance to declare convergence.

    max_iter : integer, default 100
        The maximum number of iterations.

    verbose : boolean, default False
        If verbose is True, the objective function, rnorm and snorm are
        printed at each iteration.

    assume_centered : boolean, default False
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    time_on_axis : {'first', 'last'}, default 'first'
        If data have time as the last dimension, set this to 'last'.
        Useful to use scikit-learn functions as train_test_split.

    update_rho_options : dict, default None
        Options for the update of rho. See `update_rho` function for details.

    compute_objective : boolean, default True
        Choose if compute the objective function during iterations
        (only useful if `verbose=True`).

    mode : {'admm'}, default 'admm'
        Minimisation algorithm. At the moment, only 'admm' is available,
        so this is ignored.

    Attributes
    ----------
    covariance_ : array-like, shape (n_times, n_features, n_features)
        Estimated covariance matrix

    precision_ : array-like, shape (n_times, n_features, n_features)
        Estimated pseudo inverse matrix.

    latent_ : array-like, shape (n_times, n_features, n_features)
        Estimated latent variable matrix.

    n_iter_ : int
        Number of iterations run.

    """

    def __init__(
            self, alpha=0.01, tau=1., beta=1., eta=1., mode='admm', rho=1.,
            time_on_axis='first', tol=1e-4, rtol=1e-4, psi='laplacian',
            phi='laplacian', max_iter=100, verbose=False,
            assume_centered=False, update_rho_options=None,
            compute_objective=True):
        super(LatentTimeGraphLasso, self).__init__(
            alpha=alpha, beta=beta, mode=mode, rho=rho, tol=tol, rtol=rtol,
            psi=psi, max_iter=max_iter, verbose=verbose,
            time_on_axis=time_on_axis, assume_centered=assume_centered,
            update_rho_options=update_rho_options,
            compute_objective=compute_objective)
        self.tau = tau
        self.eta = eta
        self.phi = phi

    def get_observed_precision(self):
        """Getter for the observed precision matrix.

        Returns
        -------
        precision_ : array-like,
            The precision matrix associated to the current covariance object.
            Note that this is the observed precision matrix.

        """
        return self.precision_ - self.latent_

    def _fit(self, emp_cov, n_samples):
        """Fit the LatentTimeGraphLasso model to X.

        Parameters
        ----------
        emp_cov : ndarray, shape (n_features, n_features)
            Empirical covariance of data.

        """
        self.precision_, self.latent_, self.covariance_, self.n_iter_ = \
            latent_time_graph_lasso(
                emp_cov, n_samples=n_samples,
                alpha=self.alpha, tau=self.tau, rho=self.rho,
                beta=self.beta, eta=self.eta, mode=self.mode,
                tol=self.tol, rtol=self.rtol, psi=self.psi, phi=self.phi,
                max_iter=self.max_iter, verbose=self.verbose,
                return_n_iter=True, return_history=False,
                update_rho_options=self.update_rho_options,
                compute_objective=self.compute_objective)
        return self



#!/usr/bin/python
"""regain setup script.

Author: Federico Tomasi
Copyright (c) 2017, Federico Tomasi.
Licensed under the BSD 3-Clause License (see LICENSE.txt).
"""

from setuptools import setup, find_packages
# import numpy as np

# Package Version
from regain import __version__ as version
setup(
    name='regain',
    version=version,

    description=('REGAIN (Regularised Graph Inference)'),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Federico Tomasi',
    author_email='federico.tomasi@dibris.unige.it',
    maintainer='Federico Tomasi',
    maintainer_email='federico.tomasi@dibris.unige.it',
    url='https://github.com/fdtomasi/regain',
    download_url='https://github.com/fdtomasi/regain/archive/'
                 'v%s.tar.gz' % version,
    keywords=['graph inference', 'latent variables'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python'
    ],
    license='FreeBSD',
    packages=find_packages(exclude=["*.__old", "*.tests"]),
    include_package_data=True,
    requires=['numpy (>=1.11)',
              'scipy (>=0.16.1,>=1.0)',
              'sklearn (>=0.17)',
              'six'],
    # scripts=['scripts/ici_run.py', 'scripts/ici_analysis.py'],
    # ext_modules=[ssk_module],
    # include_dirs=[np.get_include()]
)



"""Graphical latent variable model selection via ADMM."""
from __future__ import division

import warnings

import numpy as np
from scipy import linalg
from six.moves import range

from regain.covariance.graph_lasso_ import GraphLasso, logl
from regain.norm import l1_od_norm
from regain.prox import prox_logdet, prox_trace_indicator, soft_thresholding
from regain.update_rules import update_rho
from regain.utils import convergence


def objective(emp_cov, R, K, L, alpha, tau):
    """Objective function for latent graphical lasso."""
    obj = - logl(emp_cov, R)
    obj += alpha * l1_od_norm(K)
    obj += tau * np.linalg.norm(L, ord='nuc')
    return obj


def latent_graph_lasso(
        emp_cov, alpha=1., tau=1., rho=1., max_iter=100,
        verbose=False, tol=1e-4, rtol=1e-2, return_history=False,
        return_n_iter=True, mode=None,
        update_rho_options=None, compute_objective=True):
    r"""Latent variable graphical lasso solver.

    Solves the following problem via ADMM:
        min - log_likelihood(S, K-L) + alpha ||K||_{od,1} + tau ||L_i||_*

    where S is the empirical covariance of the data
    matrix D (training observations by features).

    Parameters
    ----------
    emp_cov : array-like
        Empirical covariance matrix.
    alpha, tau : float, optional
        Regularisation parameters.
    rho : float, optional
        Augmented Lagrangian parameter.
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Absolute tolerance for convergence.
    rtol : float, optional
        Relative tolerance for convergence.
    return_history : bool, optional
        Return the history of computed values.
    return_n_iter : bool, optional
        Return the number of iteration before convergence.
    verbose : bool, default False
        Print info at each iteration.

    Returns
    -------
    K, L : np.array, 2-dimensional, size (d x d)
        Solution to the problem.
    S : np.array, 2 dimensional
        Empirical covariance matrix.
    n_iter : int
        If return_n_iter, returns the number of iterations before convergence.
    history : list
        If return_history, then also a structure that contains the
        objective value, the primal and dual residual norms, and tolerances
        for the primal and dual residual norms at each iteration.

    """
    # _, n_features = emp_cov.shape
    # covariance_ = emp_cov.copy()
    # covariance_ *= 0.95
    # covariance_.flat[::n_features + 1] = emp_cov.flat[::n_features + 1]
    # K = linalg.pinvh(covariance_)

    K = np.zeros_like(emp_cov)
    L = np.zeros_like(emp_cov)
    U = np.zeros_like(emp_cov)
    R_old = np.zeros_like(emp_cov)

    checks = []
    for iteration_ in range(max_iter):
        # update R
        A = K - L - U
        A += A.T
        A /= 2.
        R = prox_logdet(emp_cov - rho * A, lamda=1. / rho)

        A = L + R + U
        K = soft_thresholding(A, lamda=alpha / rho)

        A = K - R - U
        A += A.T
        A /= 2.
        L = prox_trace_indicator(A, lamda=tau / rho)

        # update residuals
        U += R - K + L

        # diagnostics, reporting, termination checks
        obj = objective(emp_cov, R, K, L, alpha, tau) \
            if compute_objective else np.nan
        rnorm = np.linalg.norm(R - K + L)
        snorm = rho * np.linalg.norm(R - R_old)
        check = convergence(
            obj=obj, rnorm=rnorm, snorm=snorm,
            e_pri=np.sqrt(R.size) * tol + rtol * max(
                np.linalg.norm(R), np.linalg.norm(K - L)),
            e_dual=np.sqrt(R.size) * tol + rtol * rho * np.linalg.norm(U)
        )
        R_old = R.copy()

        if verbose:
            print("obj: %.4f, rnorm: %.4f, snorm: %.4f,"
                  "eps_pri: %.4f, eps_dual: %.4f" % check[:5])

        checks.append(check)
        if check.rnorm <= check.e_pri and check.snorm <= check.e_dual:
            break
        if check.obj == np.inf:
            break
        rho_new = update_rho(rho, rnorm, snorm, iteration=iteration_,
                             **(update_rho_options or {}))
        # scaled dual variables should be also rescaled
        U *= rho / rho_new
        rho = rho_new
    else:
        warnings.warn("Objective did not converge.")

    covariance_ = linalg.pinvh(K)
    return_list = [K, L, covariance_]
    if return_history:
        return_list.append(checks)
    if return_n_iter:
        return_list.append(iteration_)
    return return_list


class LatentGraphLasso(GraphLasso):
    """Sparse inverse covariance estimation with an l1-penalized estimator.

    Parameters
    ----------
    alpha : positive float, default 0.01
        Regularization parameter for precision matrix. The higher alpha,
        the more regularization, the sparser the inverse covariance.

    tau : positive float, default 1
        Regularization parameter for latent variables matrix. The higher tau,
        the more regularization, the lower rank of the latent matrix.

    rho : positive float, default 1
        Augmented Lagrangian parameter.

    over_relax : positive float, deafult 1
        Over-relaxation parameter (typically between 1.0 and 1.8).

    tol : positive float, default 1e-4
        Absolute tolerance to declare convergence.

    rtol : positive float, default 1e-4
        Relative tolerance to declare convergence.

    max_iter : integer, default 100
        The maximum number of iterations.

    verbose : boolean, default False
        If verbose is True, the objective function, rnorm and snorm are
        printed at each iteration.

    assume_centered : boolean, default False
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    update_rho_options : dict, default None
        Options for the update of rho. See `update_rho` function for details.

    compute_objective : boolean, default True
        Choose if compute the objective function during iterations
        (only useful if `verbose=True`).

    mode : {'admm'}, default 'admm'
        Minimisation algorithm. At the moment, only 'admm' is available,
        so this is ignored.

    Attributes
    ----------
    covariance_ : array-like, shape (n_features, n_features)
        Estimated covariance matrix

    precision_ : array-like, shape (n_features, n_features)
        Estimated pseudo inverse matrix.

    latent_ : array-like, shape (n_features, n_features)
        Estimated latent variable matrix.

    n_iter_ : int
        Number of iterations run.

    """

    def __init__(self, alpha=0.01, tau=1., rho=1., tol=1e-4, rtol=1e-4,
                 max_iter=100, verbose=False, assume_centered=False,
                 mode='admm', update_rho_options=None, compute_objective=True):
        super(LatentGraphLasso, self).__init__(
            alpha=alpha, rho=rho,
            tol=tol, rtol=rtol, max_iter=max_iter, verbose=verbose,
            assume_centered=assume_centered, mode=mode,
            update_rho_options=update_rho_options,
            compute_objective=compute_objective)
        self.tau = tau

    def get_precision(self):
        """Getter for the precision matrix.

        Returns
        -------
        precision_ : array-like,
            The precision matrix associated to the current covariance object.
            Note that this is the observed precision matrix.

        """
        return self.precision_ - self.latent_

    def _fit(self, emp_cov):
        """Fit the LatentGraphLasso model to X.

        Parameters
        ----------
        emp_cov : ndarray, shape (n_features, n_features)
            Empirical covariance of data.

        """
        self.precision_, self.latent_, self.covariance_, self.n_iter_ = \
            latent_graph_lasso(
                emp_cov, alpha=self.alpha, tau=self.tau, rho=self.rho,
                mode=self.mode, tol=self.tol, rtol=self.rtol,
                max_iter=self.max_iter, verbose=self.verbose,
                return_n_iter=True, return_history=False,
                update_rho_options=self.update_rho_options,
                compute_objective=self.compute_objective)
        return self



"""PASPAL implementation (in Python, based on the original MATLAB)."""
from __future__ import print_function, division

from itertools import combinations

import numpy as np
from scipy import linalg


def glo_prox(w0, tau, blocks, weights, lamda0, tol, max_iter, verbose=0):
    """Proximity operator of the group lasso with overlap penalty.

    First, identifies "active" blocks, then apply Bersekas's
    projected Newton method on the dual space.

    Copyright 2009-2010 Sofia Mosci and Lorenzo Rosasco

    """
    d = w0.size
    B = len(blocks)

    # ------------NEW
    # weights = zeros(B,1);
    # for g = 1:B;
    #     weights(g) = length(blocks{g})*tau^2; %weight is sqrt{|g|}
    #     weights(g) = tau^2; %not weighted
    # end
    # ------------NEW
    weights = (weights * tau) ** 2

    # if lamda is not initialized, then initialize it to 0
    if lamda0 == []:
        lamda0 = np.zeros(B)

    beta = .5
    sigma = .1
    s_beta = 1
    epsilon = 0.001

    lamda_tot = np.zeros(B)

    # Identify active blocks, by removing blocks such that w0 is already
    # inside the corresponding cylinder
    # to_be_projected = np.zeros(B, dtype=bool)
    # for g in range(B):
    #     to_be_projected[g] = np.linalg.norm(w0[blocks[g]], 2) >= weights[g]
    norm_blocks = np.array([np.linalg.norm(w0[block], 2) for block in blocks])
    to_be_projected = norm_blocks >= weights

    weights = weights[to_be_projected]
    blocks = blocks[to_be_projected]
    lamda = lamda0[to_be_projected]

    B = len(blocks)
    if B == 0:
        return np.zeros(d), 0, lamda_tot

    I = np.zeros((d, B))
    for g in range(B):
        I[blocks[g], g] = 1

    # Bersekas constrained Newton method
    i_null = 0
    for iteration_ in range(max_iter):
        if verbose > 1:
            print(".... interior, it %d out of %d" % (iteration_, max_iter))
        lamda_prev = lamda.copy()
        s = I.dot(lamda_prev)
        denominator = 1. / (1 + s)
        grad = weights - I.T.dot((w0 * denominator)**2)
        epsk = min(epsilon, np.linalg.norm(lamda - np.maximum(0, lamda - grad)))
        tmp = 2 * w0**2 * denominator**3

        I_inactive = np.where(np.logical_or(grad <= 0, lamda > epsk))[0]
        n_inactive = I_inactive.size
        B_inactive = np.zeros((n_inactive, n_inactive))

        # for g, g_inactive in enumerate(I_inactive):
        #     B_inactive[g, g] = np.sum(tmp[blocks[g_inactive]])
        #     for k in range(g + 1, n_inactive):
        #         B_inactive[k, g] = B_inactive[g, k] = np.sum(tmp[
        #             np.intersect1d(blocks[g_inactive],
        #                            blocks[I_inactive[k]])])
        # import time
        # tic = time.time()

        blocks_inactive_set = [set(blocks[c]) for c in I_inactive]
        B_inactive[np.triu_indices_from(B_inactive, 1)] = [
            tmp[list(x.intersection(y))].sum()
            for x, y in combinations(blocks_inactive_set, 2)]
        B_inactive += B_inactive.T
        B_inactive.flat[::n_inactive + 1] = [
            tmp[list(x)].sum() for x in blocks_inactive_set]

        # print("time 1: ", time.time() - tic)
        # tic = time.time()
        #
        # blocks_inactive = [blocks[c] for c in I_inactive]
        # C_inactive[np.triu_indices_from(C_inactive, 1)] = [
        #     tmp[np.intersect1d(x, y)].sum()
        #     for x, y in combinations(blocks_inactive, 2)]
        # C_inactive += C_inactive.T
        # C_inactive.flat[::n_inactive + 1] = [
        #     tmp[x].sum() for x in blocks_inactive]
        #
        # print("time 2: ", time.time() - tic)
        # assert np.allclose(B_inactive, C_inactive)

        p_inactive = linalg.pinv(B_inactive).dot(grad[I_inactive])

        I_active = np.where(np.logical_and(grad > 0, lamda <= epsk))[0]
        # n_active = I_active.size
        if n_inactive == 0:
            i_null += 1
        else:
            i_null = 0

        if i_null >= 5:
            break

        # B_active = np.empty(n_active)
        B_active = np.array(
            [tmp[blocks[g_active]].sum() for g_active in I_active])

        p_active = grad[I_active] / B_active

        x_inactive = grad[I_inactive].T.dot(p_inactive)
        test = 1
        m = 0
        lamda_m = np.zeros(B)
        while test:
            m += 1
            step = beta**m * s_beta
            lamda_m[I_active] = np.maximum(
                0, lamda_prev[I_active] - step * p_active)
            lamda_m[I_inactive] = np.maximum(
                0, lamda_prev[I_inactive] - step * p_inactive)
            s_m = I.dot(lamda_m)
            fdiff = (w0**2).T.dot(I.dot(lamda_m - lamda) / ((1+s_m)*(1+s))) \
                + np.sum(weights * (lamda - lamda_m))
            x_active = grad[I_active].T.dot(lamda[I_active]-lamda_m[I_active])
            test = fdiff < sigma * (step * x_inactive + x_active)

        lamda = lamda_m.copy()
        if all(grad[lamda == 0] >= 0) and all(abs(grad[lamda > 0]) < tol):
            break

    # given the solution of dual problem, lamda, compute the primal solution w
    s = I.dot(lamda)
    w = w0 * (1 - 1. / (1 + s))
    lamda_tot[to_be_projected] = lamda

    return w, iteration_, lamda_tot


def glopridu_algorithm(
    X, Y, blocks, tau, weights=None, smooth_par=0, beta0=None, lamda0=None,
    sigma0=None, max_iter_ext=1e4, max_iter_int=100, tol_ext=1e-6,
        tol_int=1e-4, verbose=0):
    """Blabla."""
    n, d = X.shape
    blocks = np.array(blocks)

    # if only the number of blocks is given, build blocks of equal
    # cardinality with sequential features
    # if ~iscell(blocks);
    #     blocks = num2cell(reshape(1:d,d/blocks,blocks),1);
    # end

    # if sigma is not specified in input, set  it to as a/n
    if not sigma0:
        sigma0 = np.linalg.norm(X, 2) ** 2 / n  # step size for smooth_par=0

    # if weights are not specified in input, set them to 1
    if not weights:
        weights = np.ones(len(blocks))
        # weights = np.ones(d)

    mu = smooth_par * sigma0  # smoothness parameter is rescaled
    sigma = sigma0 + mu  # step size

    # useful normalization that avoid computing the same computations for
    # each iteration
    mu_s = mu / sigma
    tau_s = tau / sigma
    XT = X.T / (n * sigma)

    # initialization
    if not beta0:
        beta0 = np.zeros(d)

    beta = beta0.copy()  # initialization for iterate n_iter-1
    h = beta0.copy()  # initialization for combination of the previous 2 iterates (iteratations n_iter_1 and n_iter-2)
    t = 1  # initialization for the adaptive parameter used to combine the previous 2 iterates when building h
    # precomputes X*beta and X*h to avoid computing them twice
    Xb = X.dot(beta)
    Xh = Xb.copy()
    if lamda0 is None:
        lamda0 = np.zeros(len(blocks))

    # initialization for the dual vector in computing the projection
    lamda_prev = lamda0.copy()

    for n_iter in range(1, int(max_iter_ext) + 1):
        if verbose > 0:
            print("Iteration %d/%d" % (n_iter - 1, max_iter_ext))

        beta_prev = beta.copy()
        Xb_prev = Xb.copy()

        # computes the gradient step
        #beta_noproj = h.*(1-mu_s) + XT*(Y-Xh);

        # compute the proximity operator with tolerance depending on k
        beta, q, lamda = glo_prox(
            h * (1. - mu_s) + XT.dot(Y - Xh), tau_s, blocks, weights,
            lamda_prev, tol_int * n_iter**(-3. / 2), max_iter_int,
            verbose=verbose)

        lamda_prev = lamda.copy()
        Xb = X.dot(beta)

        t_new = .5 * (1 + np.sqrt(1 + 4 * t * t))

        # combination of the 2 previous iterates
        h = beta + (t - 1) / t_new * (beta - beta_prev)
        Xh = Xb * (1. + (t - 1.) / t_new) + (1. - t) / t_new * Xb_prev
        t = t_new

        if np.linalg.norm(Xb - Xb_prev) <= np.linalg.norm(Xb_prev) * tol_ext \
                and n_iter > 1:
            break

    return beta, lamda, n_iter
